"""EULA TUI spoke for the Initial Setup"""

import codecs
import logging

from pyanaconda.ui.tui.spokes import NormalTUISpoke
from simpleline.render.widgets import TextWidget, CheckboxWidget, SeparatorWidget
from simpleline.render.containers import ListRowContainer
from simpleline.render.screen import UIScreen, InputState
from simpleline.render.screen_handler import ScreenHandler
from pyanaconda.ui.common import FirstbootOnlySpokeMixIn
from initial_setup.product import get_license_file_name, eula_available
from initial_setup.common import LicensingCategory
from initial_setup.i18n import _, N_
from pykickstart.constants import FIRSTBOOT_RECONFIG

log = logging.getLogger("initial-setup")


__all__ = ["EULASpoke"]


class EULASpoke(FirstbootOnlySpokeMixIn, NormalTUISpoke):
    """The EULA spoke providing ways to read the license and agree/disagree with it."""

    category = LicensingCategory

    def __init__(self, *args, **kwargs):
        NormalTUISpoke.__init__(self, *args, **kwargs)
        self.title = _("License information")
        self._container = None

    def initialize(self):
        NormalTUISpoke.initialize(self)

    def refresh(self, args=None):
        NormalTUISpoke.refresh(self, args)

        self._container = ListRowContainer(1)

        log.debug("license found")
        # make the options aligned to the same column (the checkbox has the
        # '[ ]' prepended)
        self._container.add(TextWidget("%s\n" % _("Read the License Agreement")),
                            self._show_license_screen_callback)

        self._container.add(CheckboxWidget(title=_("I accept the license agreement."),
                                           completed=self.data.eula.agreed),
                            self._license_accepted_callback)
        self.window.add_with_separator(self._container)

    @property
    def completed(self):
        # Either there is no EULA available, or user agrees/disagrees with it.
        return self.data.eula.agreed

    @property
    def mandatory(self):
        # This spoke is always mandatory.
        return True

    @property
    def status(self):
        return _("License accepted") if self.data.eula.agreed else _("License not accepted")

    @classmethod
    def should_run(cls, environment, data):
        # the EULA spoke should only run in Initial Setup
        if eula_available() and FirstbootOnlySpokeMixIn.should_run(environment, data):
            # don't run if we are in reconfig mode and the EULA has already been accepted
            if data and data.firstboot.firstboot == FIRSTBOOT_RECONFIG and data.eula.agreed:
                log.debug("not running license spoke: reconfig mode & license already accepted")
                return False
            return True
        return False

    def apply(self):
        # nothing needed here, the agreed field is changed in the input method
        pass

    def input(self, args, key):
        if not self._container.process_user_input(key):
            return key

        return InputState.PROCESSED

    @staticmethod
    def _show_license_screen_callback(data):
        # show license
        log.debug("showing the license")
        eula_screen = LicenseScreen()
        ScreenHandler.push_screen(eula_screen)

    def _license_accepted_callback(self, data):
        # toggle EULA agreed checkbox by changing ksdata
        log.debug("license accepted state changed to: %s", self.data.eula.agreed)
        self.data.eula.agreed = not self.data.eula.agreed
        self.redraw()


class LicenseScreen(UIScreen):
    """Screen showing the License without any input from user requested."""

    def __init__(self):
        super().__init__()

        self._license_file = get_license_file_name()

    def refresh(self, args=None):
        super().refresh(args)

        # read the license file and make it one long string so that it can be
        # processed by the TextWidget to fit in the screen in a best possible
        # way
        log.debug("reading the license file")
        with open(self._license_file, 'r') as f:
            license_text = f.read()

        self.window.add_with_separator(TextWidget(license_text))

    def input(self, args, key):
        """ Handle user input. """
        return InputState.PROCESSED_AND_CLOSE

    def prompt(self, args=None):
        # we don't want to prompt user, just close the screen
        self.close()
        return None



"""EULA spoke for the Initial Setup"""

import logging

from pyanaconda.ui.common import FirstbootOnlySpokeMixIn
from pyanaconda.ui.gui.spokes import NormalSpoke
from pykickstart.constants import FIRSTBOOT_RECONFIG

from initial_setup.product import eula_available, get_license_file_name
from initial_setup.common import LicensingCategory
from initial_setup.i18n import _, N_

log = logging.getLogger("initial-setup")


__all__ = ["EULASpoke"]

class EULASpoke(FirstbootOnlySpokeMixIn, NormalSpoke):
    """The EULA spoke"""

    builderObjects = ["eulaBuffer", "eulaWindow"]
    mainWidgetName = "eulaWindow"
    uiFile = "eula.glade"
    # The EULA spoke is self-explanatory, so just redirect its
    # help button to the Initial Setup Hub help file,
    # which covers the overall theory of Initial Setup
    # usage.
    helpFile = "InitialSetupHub.xml"

    icon = "application-certificate-symbolic"
    title = N_("_License Information")
    category = LicensingCategory
    translationDomain = "initial-setup"

    def initialize(self):
        log.debug("initializing the EULA spoke")
        NormalSpoke.initialize(self)

        self._have_eula = True
        self._eula_buffer = self.builder.get_object("eulaBuffer")
        self._agree_check_button = self.builder.get_object("agreeCheckButton")
        self._agree_label = self._agree_check_button.get_child()
        self._agree_text = self._agree_label.get_text()

        log.debug("looking for the license file")
        license_file = get_license_file_name()
        if not license_file:
            log.error("no license found")
            self._have_eula = False
            self._eula_buffer.set_text(_("No license found. Please report this "
                                         "at http://bugzilla.redhat.com"))
            return

        self._eula_buffer.set_text("")
        itr = self._eula_buffer.get_iter_at_offset(0)
        log.debug("opening the license file")
        with open(license_file, "r") as fobj:
            # insert the first line without prefixing with space
            try:
                first_line = next(fobj)
            except StopIteration:
                # nothing in the file
                return
            self._eula_buffer.insert(itr, first_line.strip())

            # EULA file is preformatted for the console, we want to let Gtk
            # format it (blank lines should be preserved)
            for line in fobj:
                stripped_line = line.strip()
                if stripped_line:
                    self._eula_buffer.insert(itr, " " + stripped_line)
                else:
                    self._eula_buffer.insert(itr, "\n\n")

    def refresh(self):
        self._agree_check_button.set_sensitive(self._have_eula)
        self._agree_check_button.set_active(self.data.eula.agreed)

    def apply(self):
        self.data.eula.agreed = self._agree_check_button.get_active()

    @property
    def completed(self):
        return not self._have_eula or self.data.eula.agreed

    @property
    def status(self):
        if not self._have_eula:
            return _("No license found")

        return _("License accepted") if self.data.eula.agreed else _("License not accepted")

    @classmethod
    def should_run(cls, environment, data):
        # the EULA spoke should only run if both are true:
        # - we are running in Initial Setup
        # - an EULA is available
        if eula_available() and FirstbootOnlySpokeMixIn.should_run(environment, data):
            # don't run if we are in reconfig mode and the EULA has already been accepted
            if data and data.firstboot.firstboot == FIRSTBOOT_RECONFIG and data.eula.agreed:
                log.debug("not running license spoke: reconfig mode & license already accepted")
                return False
            return True
        return False

    def on_check_button_toggled(self, checkbutton, *args):
        if self._agree_check_button.get_active():
            log.debug("license is now accepted")
            self._agree_label.set_markup("<b>%s</b>" % self._agree_text)
        else:
            log.debug("license no longer accepted")
            self._agree_label.set_markup(self._agree_text)



"""Common methods for Initial Setup"""

import os

from pyanaconda.ui.common import collect
from pyanaconda.core.constants import FIRSTBOOT_ENVIRON
from pyanaconda.core.i18n import N_
from pyanaconda.ui.categories import SpokeCategory

from initial_setup.product import eula_available

# a set of excluded console names
# - console, tty, tty0 -> these appear to be just aliases to the  default console,
#                         leaving them in would result in duplicate input on and output from
#                         the default console
# - ptmx -> pseudoterminal configuration device not intended as general user-facing console that
#           starts to block on write after some amount of characters has been written to it
# - ttyUSB0-4 -> used by some GSM modems apparently, blocks when TUI tries to use it,
#                 see rhbz#1755580 for more details about the hardware in question

TUI_EXCLUDED_CONSOLES = {"console", "tty", "tty0", "ptmx", "ttyUSB0", "ttyUSB1", "ttyUSB2", "ttyUSB3", "ttyUSB4"}

def collect_spokes(mask_paths, spoke_parent_class):
    """Return a list of all spoke subclasses that should appear for a given
       category. Look for them in files imported as module_path % basename(f)

       :param mask_paths: list of mask, path tuples to search for classes
       :type mask_paths: list of (mask, path)

       :param spoke_parent_class: Spoke parent class used for checking spoke compatibility
       :type spoke_parent_class: GUI or TUI Spoke class

       :return: list of Spoke classes belonging to category
       :rtype: list of Spoke classes

    """
    spokes = []
    for mask, path in mask_paths:

        spokes.extend(collect(mask, path,
                              lambda obj: issubclass(obj, spoke_parent_class) and obj.should_run("firstboot", None)))
    return spokes


def collectCategoriesAndSpokes(hub_instance, spoke_parent_class):
    """Collects categories and spokes to be displayed on this Hub,
       this method overrides the Anacondas implementation so that
       spokes relevant to Initial setup are collected

       :param hub_instance: an Initial Setup GUI or TUI Hub class instance
       :type hub_instance: class instance

       :param spoke_parent_class: Spoke parent class used for checking spoke compatibility
       :type spoke_parent_class: GUI or TUI Spoke class

       :return: dictionary mapping category class to list of spoke classes
       :rtype: dictionary[category class] -> [ list of spoke classes ]
    """
    ret = {}

    # Collect all the categories this hub displays, then collect all the
    # spokes belonging to all those categories.
    candidate_spokes = collect_spokes(hub_instance.paths["spokes"], spoke_parent_class)
    spokes = [spoke for spoke in candidate_spokes
              if spoke.should_run(FIRSTBOOT_ENVIRON, hub_instance.data)]

    for spoke in spokes:
        ret.setdefault(spoke.category, [])
        ret[spoke.category].append(spoke)

    return ret

def console_filter(console_name):
    """Filter out consoles we don't want to attempt running the TUI on.

    This at the moment just means console aliases, but it's possible more
    consoles will have to be added for other reasons in the guture.

    :param str console_name: console name to check
    :returns: if the console name is considered usable for IS TUI
    :rtype: bool
    """
    return console_name not in TUI_EXCLUDED_CONSOLES

def list_usable_consoles_for_tui():
    """List suitable consoles for running the Initial Setup TUI.

    We basically want to list any console a user might using,
    as we can't really be sure which console is in use or not.

    :returns: a list of console names considered usable for the IS TUI
    :rtype: list
    """
    console_names = [c for c in os.listdir("/sys/class/tty/") if console_filter(c)]
    return sorted(console_names)

def parse_os_release_lines(osreleaselines):
    """man os-release for format

       :return: dictionary mapping of line
       :rtype: dictionary
    """
    osrel = {}

    for line in osreleaselines:
        line = line.strip()
        try:
            key, sep, value = line.parition('=')
            osrel[key] = value.strip('"')
        except ValueError:
            # line.parition returned too many or too few items
            # in either case do nothing special
            pass

    return osrel

def parse_os_release_file(filename='/etc/os-release'):
    """Read os-release into a dictionary

       :return: dictionary mapping of os-release
       :rtype: dictionary
    """
    osrel = {}
    with open(filename) as osrelfil:
        osrel = parse_os_release_lines(osrelfil)

    return osrel

def os_bug_report_url(filename='/etc/os-release'):
    """Read os-release and find the BUG_REPORT_URL

       :return: url in (hopefully) RFC3986 format
       :rtype: string or None
    """
    osrel = parse_os_release_file(filename)
    return osrel.get('BUG_REPORT_URL')

def get_quit_message():
    if eula_available():
        return N_("Are you sure you want to quit the configuration process?\n"
                  "You might end up with an unusable system if you do. Unless the "
                  "License agreement is accepted, the system will be rebooted.")
    else:
        return N_("Are you sure you want to quit the configuration process?\n"
                  "You might end up with unusable system if you do.")

class LicensingCategory(SpokeCategory):
    displayOnHubGUI = "ProgressHub"
    displayOnHubTUI = "SummaryHub"
    sortOrder = 100
    title = N_("LICENSING")



from pyanaconda.ui.gui import QuitDialog, GraphicalUserInterface
from initial_setup.product import product_title, is_final
from initial_setup.common import get_quit_message
from initial_setup.i18n import _, N_
from .hubs import InitialSetupMainHub
import os
from gi.repository import Gdk

class InitialSetupQuitDialog(QuitDialog):
    MESSAGE = get_quit_message()

class InitialSetupGraphicalUserInterface(GraphicalUserInterface):
    """This is the main Gtk based firstboot interface. It inherits from
       anaconda to make the look & feel as similar as possible.
    """

    screenshots_directory = "/tmp/initial-setup-screenshots"

    def __init__(self):
        GraphicalUserInterface.__init__(self, None, None, product_title, is_final,
                                        quitDialog=InitialSetupQuitDialog)
        self.mainWindow.set_title("")

    def _list_hubs(self):
        return [InitialSetupMainHub]

    basemask = "initial_setup.gui"
    basepath = os.path.dirname(__file__)
    paths = GraphicalUserInterface.paths + {
        "spokes": [(basemask + ".spokes.%s", os.path.join(basepath, "spokes"))],
        "categories": [(basemask + ".categories.%s", os.path.join(basepath, "categories"))],
        }




# Pylint checker for pointless class attributes overrides.
#
# Copyright (C) 2014  Red Hat, Inc.
#
# This copyrighted material is made available to anyone wishing to use,
# modify, copy, or redistribute it subject to the terms and conditions of
# the GNU General Public License v.2, or (at your option) any later version.
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY expressed or implied, including the implied warranties of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General
# Public License for more details.  You should have received a copy of the
# GNU General Public License along with this program; if not, write to the
# Free Software Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA
# 02110-1301, USA.  Any Red Hat trademarks that are incorporated in the
# source code or documentation are not subject to the GNU General Public
# License and may only be used or replicated with the express permission of
# Red Hat, Inc.
#
# Red Hat Author(s): Anne Mulhern <amulhern@redhat.com>
#

import abc

from six import add_metaclass

import astroid

from pylint.checkers import BaseChecker
from pylint.checkers.utils import check_messages
from pylint.interfaces import IAstroidChecker

@add_metaclass(abc.ABCMeta)
class PointlessData(object):

    _DEF_CLASS = abc.abstractproperty(doc="Class of interesting definitions.")
    message_id = abc.abstractproperty(doc="Pylint message identifier.")

    @classmethod
    @abc.abstractmethod
    def _retain_node(cls, node, restrict=True):
        """ Determines whether to retain a node for the analysis.

            :param node: an AST node
            :type node: astroid.Class
            :param restrict bool: True if results returned should be restricted
            :returns: True if the node should be kept, otherwise False
            :rtype: bool

            Restricted nodes are candidates for being marked as overridden.
            Only restricted nodes are put into the initial pool of candidates.
        """
        raise NotImplementedError()

    @staticmethod
    @abc.abstractmethod
    def _extract_value(node):
        """ Return the node that contains the assignment's value.

            :param node: an AST node
            :type node: astroid.Class
            :returns: the node corresponding to the value
            :rtype: bool
        """
        raise NotImplementedError()

    @staticmethod
    @abc.abstractmethod
    def _extract_targets(node):
        """ Generates the names being assigned to.

            :param node: an AST node
            :type node: astroid.Class
            :returns: a list of assignment target names
            :rtype: generator of str
        """
        raise NotImplementedError()

    @classmethod
    def get_data(cls, node, restrict=True):
        """ Find relevant nodes for this analysis.

            :param node: an AST node
            :type node: astroid.Class
            :param restrict bool: True if results returned should be restricted

            :rtype: generator of astroid.Class
            :returns: a generator of interesting nodes.

            Note that all nodes returned are guaranteed to be instances of
            some class in self._DEF_CLASS.
        """
        nodes = (n for n in node.body if isinstance(n, cls._DEF_CLASS))
        for n in nodes:
            if cls._retain_node(n, restrict):
                for name in cls._extract_targets(n):
                    yield (name, cls._extract_value(n))

    @classmethod
    @abc.abstractmethod
    def check_equal(cls, node, other):
        """ Check whether the two nodes are considered equal.

            :param node: some ast node
            :param other: some ast node

            :rtype: bool
            :returns: True if the nodes are considered equal, otherwise False

            If the method returns True, the nodes are actually equal, but it
            may return False when the nodes are equal.
        """
        raise NotImplementedError()

class PointlessFunctionDefinition(PointlessData):
    """ Looking for pointless function definitions. """

    _DEF_CLASS = astroid.Function
    message_id = "W9952"

    @classmethod
    def _retain_node(cls, node, restrict=True):
        return not restrict or \
           (len(node.body) == 1 and isinstance(node.body[0], astroid.Pass))

    @classmethod
    def check_equal(cls, node, other):
        return len(node.body) == 1 and isinstance(node.body[0], astroid.Pass) and \
           len(other.body) == 1 and isinstance(other.body[0], astroid.Pass)

    @staticmethod
    def _extract_value(node):
        return node

    @staticmethod
    def _extract_targets(node):
        yield node.name

class PointlessAssignment(PointlessData):

    _DEF_CLASS = astroid.Assign
    message_id = "W9951"

    _VALUE_CLASSES = (
        astroid.Const,
        astroid.Dict,
        astroid.List,
        astroid.Tuple
    )

    @classmethod
    def _retain_node(cls, node, restrict=True):
        return not restrict or isinstance(node.value, cls._VALUE_CLASSES)

    @classmethod
    def check_equal(cls, node, other):
        if type(node) != type(other):
            return False
        if isinstance(node, astroid.Const):
            return node.value == other.value
        if isinstance(node, (astroid.List, astroid.Tuple)):
            return len(node.elts) == len(other.elts) and \
               all(cls.check_equal(n, o) for (n, o) in zip(node.elts, other.elts))
        if isinstance(node, astroid.Dict):
            return len(node.items) == len(other.items)
        return False

    @staticmethod
    def _extract_value(node):
        return node.value

    @staticmethod
    def _extract_targets(node):
        for target in node.targets:
            yield target.name

class PointlessClassAttributeOverrideChecker(BaseChecker):
    """ If the nearest definition of the class attribute in the MRO assigns
        it the same value, then the overriding definition is said to be
        pointless.

        The algorithm for detecting a pointless attribute override is the following.

        * For each class, C:
           - For each attribute assignment,
               name_1 = name_2 ... name_n = l (where l is a literal):
              * For each n in (n_1, n_2):
                - Traverse the linearization of the MRO until the first
                   matching assignment n = l' is identified. If l is equal to l',
                   then consider that the assignment to l in C is a
                   pointless override.

        The algorithm for detecting a pointless method override has the same
        general structure, and the same defects discussed below.

        Note that this analysis is neither sound nor complete. It is unsound
        under multiple inheritance. Consider the following class hierarchy::

            class A(object):
                _attrib = False

            class B(A):
                _attrib = False

            class C(A):
                _attrib = True

            class D(B,C):
                pass

        In this  case, starting from B, B._attrib = False would be considered
        pointless. However, for D the MRO is B, C, A, and removing the assignment
        B._attrib = False would change the inherited value of D._attrib from
        False to True.

        The analysis is incomplete because it will find some values unequal when
        actually they are equal.

        The analysis is both incomplete and unsound because it expects that
        assignments will always be made by means of the same syntax.
    """

    __implements__ = (IAstroidChecker,)

    name = "pointless class attribute override checker"
    msgs = {
       "W9951":
       (
          "Assignment to class attribute %s overrides identical assignment in ancestor.",
          "pointless-class-attribute-override",
          "Assignment to class attribute  that overrides assignment in ancestor that assigns identical value has no effect."
       ),
       "W9952":
       (
          "definition of %s method overrides identical method definition in ancestor",
          "pointless-method-definition-override",
          "Overriding empty method definition with another empty method definition has no effect."
       )
    }

    @check_messages("W9951", "W9952")
    def visit_class(self, node):
        for checker in (PointlessAssignment, PointlessFunctionDefinition):
            for (name, value) in checker.get_data(node):
                for a in node.ancestors():
                    match = next((v for (n, v) in checker.get_data(a, False) if n == name), None)
                    if match is not None:
                        if checker.check_equal(value, match):
                            self.add_message(checker.message_id, node=value, args=(name,))
                        break

def register(linter):
    linter.register_checker(PointlessClassAttributeOverrideChecker(linter))



from pyanaconda.ui.gui.hubs import Hub
from pyanaconda.ui.gui.spokes import NormalSpoke as GUI_spoke_class
from initial_setup import product
from initial_setup import common

__all__ = ["InitialSetupMainHub"]


class InitialSetupMainHub(Hub):
    uiFile = "initial_setup.glade"
    builderObjects = ["summaryWindow"]
    mainWidgetName = "summaryWindow"
    translationDomain = "initial-setup"
    helpFile = "InitialSetupHub.xml"

    # Should we automatically go to next hub if processing is done and there are no
    # spokes on the hub ? The correct value for Initial Setup is True, due to the
    # long standing Initial Setup policy of continuing with system startup if there
    # are no spokes to be shown.
    continue_if_empty = True

    def __init__(self, *args):
        Hub.__init__(self, *args)

    def _collectCategoriesAndSpokes(self):
        return common.collectCategoriesAndSpokes(self, GUI_spoke_class)

    def _createBox(self):
        Hub._createBox(self)

        # override spokes' distribution strings set by the pyanaconda module
        for spoke in self._spokes.values():
            spoke.window.set_property("distribution",
                                      product.product_title().upper())



from pyanaconda.ui.tui import TextUserInterface
from pyanaconda import threading

from initial_setup.product import product_title, is_final
from initial_setup.common import list_usable_consoles_for_tui, get_quit_message
from initial_setup.i18n import _, N_
from .hubs import InitialSetupMainHub

from simpleline import App

import os
import sys
import select
import contextlib
import termios
import logging
log = logging.getLogger("initial-setup")

QUIT_MESSAGE = get_quit_message()

class MultipleTTYHandler(object):
    """Run the Initial Setup TUI on all usable consoles.

    This is done by redirecting the Initial Setup stdout to all
    usable consoles and then redirecting any input back to
    the Initial Setup stdin.
    """

    def __init__(self, tui_stdout_fd, tui_stdin_fd):
        # create file objects for the TUI stdout and stdin fds
        self._tui_stdout_fd = tui_stdout_fd
        self._tui_stdout = os.fdopen(tui_stdout_fd, "r")
        self._tui_stdin_fd = tui_stdin_fd
        self._tui_stdin = os.fdopen(tui_stdin_fd, "w")

        self._shutdown = False

        self._active_console = None

        self._console_read_fos = {}
        self._console_write_fos = []
        self._open_all_consoles()

    def shutdown(self):
        """Tell the multi TTY handler to shutdown."""
        self._shutdown = True

    def _open_all_consoles(self):
        """Open all consoles suitable for running the Initial Setup TUI."""
        console_write_fos = {}
        console_read_fos = {}
        console_paths = (os.path.join("/dev", c) for c in list_usable_consoles_for_tui())
        usable_console_paths = []
        unusable_console_paths = []
        for console_path in console_paths:
            try:
                write_fo = open(console_path, "w")
                read_fo = open(console_path, "r")
                fd = read_fo.fileno()
                console_write_fos[fd] = write_fo
                # the console stdin file descriptors need to be non-blocking
                os.set_blocking(fd, False)
                console_read_fos[fd] = read_fo
                # If we survived till now the console might be usable
                # (could be read and written into).
                usable_console_paths.append(console_path)
            except Exception:
                log.exception("can't open console for Initial Setup TUI: %s", console_path)
                unusable_console_paths.append(console_path)

        log.debug("The Initial Setup TUI will attempt to run on the following consoles:")
        log.debug("\n".join(usable_console_paths))
        log.debug("The following consoles could not be opened and will not be used:")
        log.debug("\n".join(unusable_console_paths))
        self._console_read_fos = console_read_fos
        self._console_write_fos = console_write_fos

    def run(self):
        """Run IS TUI on multiple consoles."""
        # we wait for data from the consoles
        fds = list(self._console_read_fos.keys())
        # as well as from the anaconda stdout
        fds.append(self._tui_stdout_fd)
        log.info("multi TTY handler starting")
        while True:
            # Watch the consoles and IS TUI stdout for data and
            # react accordingly.
            # The select also triggers every second (the 1.0 parameter),
            # so that the infinite loop can be promptly interrupted once
            # the multi TTY handler is told to shutdown.
            rlist, _wlist, _xlist = select.select(fds, [], [], 1.0)
            if self._shutdown:
                log.info("multi TTY handler shutting down")
                break
            for fd in rlist:
                if fd == self._tui_stdout_fd:
                    # We need to set the TUI stdout fd to non-blocking,
                    # as otherwise reading from it would (predictably) result in
                    # the readline() function blocking once it runs out of data.
                    os.set_blocking(fd, False)

                    # The IS TUI wants to write something,
                    # read all the lines.
                    lines = self._tui_stdout.readlines()

                    # After we finish reading all the data we need to set
                    # the TUI stdout fd to blocking again.
                    # Otherwise the fd will not be usable when we try to read from
                    # it again for unclear reasons.
                    os.set_blocking(fd, True)

                    lines.append("\n")  # seems to get lost somewhere on the way

                    # Write all the lines IS wrote to stdout to all consoles
                    # that we consider usable for the IS TUI.
                    for console_fo in self._console_write_fos.values():
                        for one_line in lines:
                            try:
                                console_fo.write(one_line)
                            except OSError:
                                log.exception("failed to write %s to console %s", one_line, console_fo)
                else:
                    # Someone typed some input to a console and hit enter,
                    # forward the input to the IS TUI stdin.
                    read_fo = self._console_read_fos[fd]
                    write_fo = self._console_write_fos[fd]
                    # as the console is getting input we consider it to be
                    # the currently active console
                    self._active_console = read_fo, write_fo
                    try:
                        data = read_fo.readline()
                    except TypeError:
                        log.exception("input reading failed for console %s", read_fo)
                        continue
                    self._tui_stdin.write(data)
                    self._tui_stdin.flush()

    def custom_getpass(self, prompt='Password: '):
        """Prompt for a password, with echo turned off that can run on an arbitrary console.

        This implementation is based on the Python 3.6 getpass() source code, with added
        support for running getpass() on an arbitrary console, as the original implementation
        is hardcoded to expect input from /dev/tty, without an option to change that.

        Raises:
          EOFError: If our input tty or stdin was closed.

        Always restores terminal settings before returning.
        """
        input_fo, output_fo = self._active_console
        passwd = None
        with contextlib.ExitStack() as stack:
            input_fd = input_fo.fileno()
            if input_fd is not None:
                try:
                    old = termios.tcgetattr(input_fd)     # a copy to save
                    new = old[:]
                    new[3] &= ~termios.ECHO  # 3 == 'lflags'
                    tcsetattr_flags = termios.TCSAFLUSH
                    if hasattr(termios, 'TCSASOFT'):
                        tcsetattr_flags |= termios.TCSASOFT
                    try:
                        termios.tcsetattr(input_fd, tcsetattr_flags, new)
                        passwd = self._raw_input(prompt, output_fo, input_fo=input_fo)
                    finally:
                        termios.tcsetattr(input_fd, tcsetattr_flags, old)
                        output_fo.flush()  # Python issue7208
                except termios.error:
                    if passwd is not None:
                        # _raw_input succeeded.  The final tcsetattr failed.  Reraise
                        # instead of leaving the terminal in an unknown state.
                        raise
                    # We can't control the tty or stdin.  Give up and use normal IO.
                    # _fallback_getpass() raises an appropriate warning.
                    if output_fo is not input_fo:
                        # clean up unused file objects before blocking
                        stack.close()
                    passwd = self._fallback_getpass(prompt, output_fo, input_fo)

            output_fo.write('\n')
            return passwd

    def _fallback_getpass(self, prompt='Password: ', output_fo=None, input_fo=None):
        log.warning("Can not control echo on the terminal: %s", input_fo)
        if not output_fo:
            output_fo = sys.stderr
        print("Warning: Password input may be echoed.", file=output_fo)
        return self._raw_input(prompt, output_fo, input_fo)


    def _raw_input(self, prompt="", output_fo=None, input_fo=None):
        # This doesn't save the string in the GNU readline history.

        # The input fd has to be set as non-blocking for the general multi-tty machinery
        # to work, but for password input to work correctly it needs to be set as blocking
        # when user input is expected.
        # We also have to switch it back to non-blocking once user input is received.
        os.set_blocking(input_fo.fileno(), True)
        prompt = str(prompt)
        if prompt:
            try:
                output_fo.write(prompt)
            except UnicodeEncodeError:
                # Use replace error handler to get as much as possible printed.
                prompt = prompt.encode(output_fo.encoding, 'replace')
                prompt = prompt.decode(output_fo.encoding)
                output_fo.write(prompt)
            output_fo.flush()
        # NOTE: The Python C API calls flockfile() (and unlock) during readline.
        line = input_fo.readline()
        if not line:
            raise EOFError
        if line[-1] == '\n':
            line = line[:-1]
        # We got input from the user, switch the input fd back to non-blocking
        # so that the multi-tty machinery works correctly.
        os.set_blocking(input_fo.fileno(), False)
        return line


class InitialSetupTextUserInterface(TextUserInterface):
    """This is the main text based firstboot interface. It inherits from
       anaconda to make the look & feel as similar as possible.
    """

    ENVIRONMENT = "firstboot"

    def __init__(self, cli_args):
        """Initialize the Initial Setup text UI.

        :param cli_args: command line arguments parsed by Argparse
        """
        TextUserInterface.__init__(self, None, None, product_title, is_final,
                                   quitMessage=QUIT_MESSAGE)

        self.multi_tty_handler = None
        self._use_multi_tty_handler = not cli_args.no_multi_tty

        # In some case, such as when running Initial Setup directly
        # in console or from an SSH session script, we should not
        # start the multi TTY handler and just run in the single
        # local console.
        if self._use_multi_tty_handler:
            # redirect stdin and stdout to custom pipes

            # stdin
            stdin_fd, tui_stdin_fd = os.pipe()
            sys.stdin = os.fdopen(stdin_fd, "r")

            # stdout
            tui_stdout_fd, stdout_fd = os.pipe()
            sys.stdout = os.fdopen(stdout_fd, "w")

            # instantiate and start the multi TTY handler
            self.multi_tty_handler = MultipleTTYHandler(tui_stdin_fd=tui_stdin_fd,
                                                        tui_stdout_fd=tui_stdout_fd)
            # start the multi-tty handler
            threading.threadMgr.add(
                threading.AnacondaThread(name="initial_setup_multi_tty_thread",
                                         target=self.multi_tty_handler.run)
            )

    def setup(self, data):
        TextUserInterface.setup(self, data)
        if self._use_multi_tty_handler:
            # Make sure custom getpass() from multi-tty handler is used instead
            # of regular getpass. This needs to be done as the default getpass()
            # implementation cant work with arbitrary consoles and always defaults
            # to /dev/tty for input.
            configuration = App.get_configuration()
            configuration.password_function = self.multi_tty_handler.custom_getpass

    def _list_hubs(self):
        return [InitialSetupMainHub]

    basemask = "firstboot.tui"
    basepath = os.path.dirname(__file__)
    paths = TextUserInterface.paths + {
        "spokes": [(basemask + ".spokes.%s", os.path.join(basepath, "spokes"))],
        "categories": [(basemask + ".categories.%s", os.path.join(basepath, "categories"))],
        }




from pyanaconda.ui.tui.hubs import TUIHub
from pyanaconda.ui.tui.spokes import NormalSpoke as TUI_spoke_class
from initial_setup import product
from initial_setup import common
from initial_setup.i18n import _, N_

__all__ = ["InitialSetupMainHub"]

class InitialSetupMainHub(TUIHub):
    categories = ["user", "localization"]

    prod_title = product.product_title()
    if prod_title:
        title = N_("Initial setup of %(product)s") % {"product": prod_title}
    else:
        title = N_("Initial setup")

    def __init__(self, *args):
        TUIHub.__init__(self, *args)

    def _collectCategoriesAndSpokes(self):
        return common.collectCategoriesAndSpokes(self, TUI_spoke_class)



__version__ = "0.3.81.5"

import os
import sys
import signal
import pykickstart
import logging
import argparse
import traceback
import atexit
from initial_setup.product import eula_available
from initial_setup import initial_setup_log
from pyanaconda.localization import setup_locale_environment, setup_locale
from pyanaconda.core.constants import FIRSTBOOT_ENVIRON, SETUP_ON_BOOT_RECONFIG, \
    SETUP_ON_BOOT_DEFAULT
from pyanaconda.flags import flags
from pyanaconda.core.startup.dbus_launcher import AnacondaDBusLauncher
from pyanaconda.modules.common.task import sync_run_task
from pyanaconda.modules.common.constants.services import BOSS, LOCALIZATION, TIMEZONE, USERS, \
    SERVICES, NETWORK
from pyanaconda.modules.common.structures.kickstart import KickstartReport


class InitialSetupError(Exception):
    pass

INPUT_KICKSTART_PATH = "/root/anaconda-ks.cfg"
OUTPUT_KICKSTART_PATH = "/root/initial-setup-ks.cfg"
RECONFIG_FILES = ["/etc/reconfigSys", "/.unconfigured"]

SUPPORTED_KICKSTART_COMMANDS = ["user",
                                "group",
                                "keyboard",
                                "lang",
                                "rootpw",
                                "timezone",
                                "logging",
                                "selinux",
                                "firewall"]

# set the environment so that spokes can behave accordingly
flags.environs = [FIRSTBOOT_ENVIRON]

signal.signal(signal.SIGINT, signal.SIG_IGN)

# setup logging
log = logging.getLogger("initial-setup")

logging_initialized = False

def log_to_journal(message, priority=3):
    """A quick-and-dirty direct Journal logger.

    A quick-and-dirty direct Journal logger used to log errors that occur
    before the normal Python logging system is setup and connected to Journal.

    :param str message: message to send to Journal
    :param int priority: message priority (2 - critical, 3 - error, 4 - warning, 5 - notice, 6 - info)
    """
    os.system('echo "%s" | systemd-cat -t initial-setup -p %s' % (message, priority))

def log_exception(*exception_info):
    exception_text = "".join(traceback.format_exception(*exception_info))
    error_message = "Initial Setup crashed due to unhandled exception:\n%s" % exception_text
    if logging_initialized:
        log.error(error_message)
    else:
        log_to_journal(error_message)

sys.excepthook = log_exception

class InitialSetup(object):
    def __init__(self, gui_mode):
        """Initialize the Initial Setup internals"""
        log.debug("initializing Initial Setup")
        # True if running in graphical mode, False otherwise (text mode)
        self.gui_mode = gui_mode
        # kickstart data
        self.data = None
        # reboot on quit flag
        self._reboot_on_quit = False

        # parse any command line arguments
        self._args = self._parse_arguments()

        # initialize logging
        initial_setup_log.init(stdout_log=not self._args.no_stdout_log)
        global logging_initialized
        logging_initialized = True

        log.info("Initial Setup %s" % __version__)

        # check if we are running as root
        if os.geteuid() != 0:
            log.critical("Initial Setup needs to be run as root")
            raise InitialSetupError

        # load configuration files
        from pyanaconda.core.configuration.anaconda import conf
        conf.set_from_files(["/etc/initial-setup/conf.d/"])

        if self.gui_mode:
            log.debug("running in GUI mode")
        else:
            log.debug("running in TUI mode")

        self._external_reconfig = False

        # check if the reconfig mode should be enabled
        # by checking if at least one of the reconfig
        # files exist
        for reconfig_file in RECONFIG_FILES:
            if os.path.exists(reconfig_file):
                self.external_reconfig = True
                log.debug("reconfig trigger file found: %s", reconfig_file)

        if self.external_reconfig:
            log.debug("running in externally triggered reconfig mode")

        if self.gui_mode:
            # We need this so we can tell GI to look for overrides objects
            # also in anaconda source directories
            import gi.overrides
            for p in os.environ.get("ANACONDA_WIDGETS_OVERRIDES", "").split(":"):
                gi.overrides.__path__.insert(0, p)
            log.debug("GI overrides imported")

        from pyanaconda.addons import collect_addon_paths

        addon_paths = ["/usr/share/initial-setup/modules", "/usr/share/anaconda/addons"]

        # append ADDON_PATHS dirs at the end
        sys.path.extend(addon_paths)

        self._addon_module_paths = collect_addon_paths(addon_paths, self.gui_mode_id)
        log.info("found %d addon modules:", len(self._addon_module_paths))
        for addon_path in self._addon_module_paths:
            log.debug(addon_path)

        # Too bad anaconda does not have modularized logging
        log.debug("initializing the Anaconda log")
        from pyanaconda import anaconda_logging
        anaconda_logging.init(write_to_journal=True)

        # create class for launching our dbus session
        self._dbus_launcher = AnacondaDBusLauncher()

        # group, user, root password set-before tracking
        self._groups_already_configured = False
        self._users_already_configured = False
        self._root_password_already_configured = False

    @property
    def external_reconfig(self):
        """External reconfig status.

        Reports if external (eq. not triggered by kickstart) has been enabled.

        :returns: True if external reconfig mode has been enabled, else False.
        :rtype: bool
        """
        return self._external_reconfig

    @external_reconfig.setter
    def external_reconfig(self, value):
        self._external_reconfig = value

    @property
    def gui_mode_id(self):
        """String id of the current GUI mode

        :returns: "gui" if gui_mode is True, "tui" otherwise
        :rtype: str
        """
        if self.gui_mode:
            return "gui"
        else:
            return "tui"

    @property
    def reboot_on_quit(self):
        # should the machine be rebooted once Initial Setup quits
        return self._reboot_on_quit

    def _parse_arguments(self):
        """Parse command line arguments"""
        # create an argparse instance
        parser = argparse.ArgumentParser(prog="Initial Setup",
                                         description="Initial Setup is can run during the first start of a newly installed"
                                         "system to configure it according to the needs of the user.")
        parser.add_argument("--no-stdout-log", action="store_true", default=False, help="don't log to stdout")
        parser.add_argument("--no-multi-tty", action="store_true", default=False,
                            help="Don't run on multiple consoles.")
        parser.add_argument('--version', action='version', version=__version__)

        # parse arguments and return the result
        return parser.parse_args()

    def _load_kickstart(self):
        """Load the kickstart"""
        from pyanaconda import kickstart

        # Construct a commandMap with only the supported Anaconda's commands
        commandMap = dict((k, kickstart.commandMap[k]) for k in SUPPORTED_KICKSTART_COMMANDS)

        # Prepare new data object
        self.data = kickstart.AnacondaKSHandler(self._addon_module_paths["ks"], commandUpdates=commandMap)

        kickstart_path = INPUT_KICKSTART_PATH
        if os.path.exists(OUTPUT_KICKSTART_PATH):
            log.info("using kickstart from previous run for input")
            kickstart_path = OUTPUT_KICKSTART_PATH

        log.info("parsing input kickstart %s", kickstart_path)
        try:
            # Read the installed kickstart
            parser = kickstart.AnacondaKSParser(self.data)
            parser.readKickstart(kickstart_path)
            log.info("kickstart parsing done")
        except pykickstart.errors.KickstartError as kserr:
            log.critical("kickstart parsing failed: %s", kserr)
            log.critical("Initial Setup startup failed due to invalid kickstart file")
            raise InitialSetupError

        # if we got this far the kickstart should be valid, so send it to Boss as well
        boss = BOSS.get_proxy()
        report = KickstartReport.from_structure(
            boss.ReadKickstartFile(kickstart_path)
        )

        if not report.is_valid():
            message = "\n\n".join(map(str, report.error_messages))
            raise InitialSetupError(message)

        if self.external_reconfig:
            # set the reconfig flag in kickstart so that
            # relevant spokes show up
            services_proxy = SERVICES.get_proxy()
            services_proxy.SetSetupOnBoot(SETUP_ON_BOOT_RECONFIG)

        # Record if groups, users or root password has been set before Initial Setup
        # has been started, so that we don't trample over existing configuration.
        users_proxy = USERS.get_proxy()
        self._groups_already_configured = bool(users_proxy.Groups)
        self._users_already_configured = bool(users_proxy.Users)
        self._root_password_already_configured = users_proxy.IsRootPasswordSet

    def _setup_locale(self):
        log.debug("setting up locale")

        localization_proxy = LOCALIZATION.get_proxy()

        # Normalize the locale environment variables
        if localization_proxy.Kickstarted:
            locale_arg = localization_proxy.Language
        else:
            locale_arg = None
        setup_locale_environment(locale_arg, prefer_environment=True)
        setup_locale(os.environ['LANG'], text_mode=not self.gui_mode)

    def _initialize_network(self):
        log.debug("initializing network")
        network_proxy = NETWORK.get_proxy()
        network_proxy.CreateDeviceConfigurations()

    def _apply(self):
        # Do not execute sections that were part of the original
        # anaconda kickstart file (== have .seen flag set)

        log.info("applying changes")

        services_proxy = SERVICES.get_proxy()
        reconfig_mode = services_proxy.SetupOnBoot == SETUP_ON_BOOT_RECONFIG

        # data.selinux
        # data.firewall

        # Configure the timezone.
        timezone_proxy = TIMEZONE.get_proxy()
        for task_path in timezone_proxy.InstallWithTasks():
            task_proxy = TIMEZONE.get_proxy(task_path)
            sync_run_task(task_proxy)

        # Configure the localization.
        localization_proxy = LOCALIZATION.get_proxy()
        for task_path in localization_proxy.InstallWithTasks():
            task_proxy = LOCALIZATION.get_proxy(task_path)
            sync_run_task(task_proxy)

        # Configure persistent hostname
        network_proxy = NETWORK.get_proxy()
        network_task = network_proxy.ConfigureHostnameWithTask(True)
        task_proxy = NETWORK.get_proxy(network_task)
        sync_run_task(task_proxy)
        # Set current hostname
        network_proxy.SetCurrentHostname(network_proxy.Hostname)

        # Configure groups, users & root account
        #
        # NOTE: We only configure groups, users & root account if the respective
        #       kickstart commands are *not* seen in the input kickstart.
        #       This basically means that we will configure only what was
        #       set in the Initial Setup UI and will not attempt to configure
        #       anything that looks like it was configured previously in
        #       the Anaconda UI or installation kickstart.
        users_proxy = USERS.get_proxy()

        if self._groups_already_configured and not reconfig_mode:
            log.debug("skipping user group configuration - already configured")
        elif users_proxy.Groups:  # only run of there are some groups to create
            groups_task = users_proxy.ConfigureGroupsWithTask()
            task_proxy = USERS.get_proxy(groups_task)
            log.debug("configuring user groups via %s task", task_proxy.Name)
            sync_run_task(task_proxy)

        if self._users_already_configured and not reconfig_mode:
            log.debug("skipping user configuration - already configured")
        elif users_proxy.Users:  # only run if there are some users to create
            users_task = users_proxy.ConfigureUsersWithTask()
            task_proxy = USERS.get_proxy(users_task)
            log.debug("configuring users via %s task", task_proxy.Name)
            sync_run_task(task_proxy)

        if self._root_password_already_configured and not reconfig_mode:
            log.debug("skipping root password configuration - already configured")
        else:
            root_task = users_proxy.SetRootPasswordWithTask()
            task_proxy = USERS.get_proxy(root_task)
            log.debug("configuring root password via %s task", task_proxy.Name)
            sync_run_task(task_proxy)

        # Configure all addons
        log.info("executing addons")
        self.data.addons.execute(storage=None, ksdata=self.data, users=None, payload=None)

        boss_proxy = BOSS.get_proxy()
        task_path = boss_proxy.InstallSystemWithTask()
        task_proxy = BOSS.get_proxy(task_path)
        sync_run_task(task_proxy)

        if self.external_reconfig:
            # prevent the reconfig flag from being written out
            # to kickstart if neither /etc/reconfigSys or /.unconfigured
            # are present
            services_proxy = SERVICES.get_proxy()
            services_proxy.SetSetupOnBoot(SETUP_ON_BOOT_DEFAULT)

        # Write the kickstart data to file
        log.info("writing the Initial Setup kickstart file %s", OUTPUT_KICKSTART_PATH)
        with open(OUTPUT_KICKSTART_PATH, "w") as f:
            f.write(str(self.data))
        log.info("finished writing the Initial Setup kickstart file")

        # Remove the reconfig files, if any - otherwise the reconfig mode
        # would start again next time the Initial Setup service is enabled.
        if self.external_reconfig:
            for reconfig_file in RECONFIG_FILES:
                if os.path.exists(reconfig_file):
                    log.debug("removing reconfig trigger file: %s" % reconfig_file)
                    os.remove(reconfig_file)

        # and we are done with applying changes
        log.info("all changes have been applied")

    def run(self):
        """Run Initial setup

        :param bool gui_mode: if GUI should be used (TUI is the default)

        :returns: True if the IS run was successful, False if it failed
        :rtype: bool
        """
        # also register boss shutdown & DBUS session cleanup via exit handler
        atexit.register(self._dbus_launcher.stop)

        # start dbus session (if not already running) and run boss in it
        try:
            self._dbus_launcher.start()
        except TimeoutError as e:
            log.error(str(e))
            return True

        self._load_kickstart()
        self._setup_locale()
        self._initialize_network()

        if self.gui_mode:
            try:
                # Try to import IS gui specifics
                log.debug("trying to import GUI")
                import initial_setup.gui
            except ImportError:
                log.exception("GUI import failed, falling back to TUI")
                self.gui_mode = False

        if self.gui_mode:
            # gui already imported (see above)

            # Add addons to search paths
            initial_setup.gui.InitialSetupGraphicalUserInterface.update_paths(self._addon_module_paths)

            # Initialize the UI
            log.debug("initializing GUI")
            ui = initial_setup.gui.InitialSetupGraphicalUserInterface()
        else:
            # Import IS gui specifics
            import initial_setup.tui

            # Add addons to search paths
            initial_setup.tui.InitialSetupTextUserInterface.update_paths(self._addon_module_paths)

            # Initialize the UI
            log.debug("initializing TUI")
            ui = initial_setup.tui.InitialSetupTextUserInterface(self._args)

        # Pass the data object to user interface
        log.debug("setting up the UI")
        ui.setup(self.data)

        # Start the application
        log.info("starting the UI")
        ret = ui.run()

        # we need to reboot the machine if there is an EULA, that was not agreed
        if eula_available() and  not self.data.eula.agreed:
            log.warning("EULA has not been agreed - the system will be rebooted.")
            self._reboot_on_quit = True

        # TUI returns False if the app was ended prematurely
        # all other cases return True or None
        if ret is False:
            log.warning("ended prematurely in TUI")
            return True

        # apply changes
        self._apply()

        # in the TUI mode shutdown the multi TTY handler (if any)
        if not self.gui_mode and ui.multi_tty_handler:
            # TODO: wait for this to finish or make it blockng ?
            ui.multi_tty_handler.shutdown()

        # and we are done
        return True
